
    use strict;
    use warnings;

    #   Configuration

    my $reader = 'John Walker';                     # Name of person who owns the list
    my $webdir = 'documents/reading_list';          # Directory on Web server where list tree resides
    my $baseurl = "http://www.fourmilab.ch/$webdir"; # URL where document tree is installed
    my $amazon_associate = 'fourmilabwwwfour';      # Amazon associate link tag
    my $abe_associate = 'click-1247935-7134924';    # Abebooks associate link tag
    my $auxwin = 'Fourmilab_readingListAux';        # Name of auxiliary window in which we open external inks

    #   End configuration

    my @books;
    my $l;
    my @months = qw(January February March April May June July August September October November December);
    my ($title, $category, $topic, $date, $author, $body) = ('', '', '', '', '', '');

    my @flattenISO = (
        ' ',       # |0xA0| Non-breaking space
        '!',       # |0xA1| Spanish open exclamation
        'cents',   # |0xA2| Cent sign
        'GBP',     # |0xA3| Pounds Sterling
        '$',       # |0xA4| Universal currency symbol
        'JPY',     # |0xA5| Japanese Yen
        '|',       # |0xA6| Broken vertical bar
        'Sec.',    # |0xA7| Section sign
        '\'\'',      # |0xA8| diaeresis
        '(C)',     # |0xA9| Copyright
        'a',       # |0xAA| Spanish feminine ordinal indicator
        '<<',      # |0xAB| Left pointing guillemet
        'NOT',     # |0xAC| Logical not
        '',        # |0xAD| Soft (discretionary) hyphen
        '(R)',     # |0xAE| Registered trademark
        '-',       # |0xAF| Overbar
        'o',       # |0xB0| Degree sign
        '+/-',     # |0xB1| Plus or minus
        '^2',      # |0xB2| Superscript 2
        '^3',      # |0xB3| Superscript 3
        '\'',      # |0xB4| Acute accent
        'mu',      # |0xB5| Micro sign
        'PP.',     # |0xB6| Paragraph sign
        '.',       # |0xB7| Middle dot
        ',',       # |0xB8| Spacing cedilla
        '^1',      # |0xB9| Superscript 1
        'o',       # |0xBA| Spanish masculine ordinal indicator
        '>>',      # |0xBB| Right pointing guillemet
        '1/4',     # |0xBC| Fraction one quarter
        '1/2',     # |0xBD| Fraction one half
        '3/4',     # |0xBE| Fraction three quarters
        '?',       # |0xBF| Spanish open question
        'A',       # |0xC0| Accented capital A grave
        'A',       # |0xC1|                    acute
        'A',       # |0xC2|                    circumflex
        'A',       # |0xC3|                    tilde
        'A',       # |0xC4|                    diaeresis
        'A',       # |0xC5| Capital A ring / Angstrom symbol
        'Ae',      # |0xC6| Capital Ae
        'C',       # |0xC7| Capital C cedilla
        'E',       # |0xC8| Accented capital E grave
        'E',       # |0xC9|                    acute
        'E',       # |0xCA|                    circumflex
        'E',       # |0xCB|                    diaeresis
        'I',       # |0xCC| Accented capital I grave
        'I',       # |0xCD|                    acute
        'I',       # |0xCE|                    circumflex
        'I',       # |0xCF|                    diaeresis
        'Th',      # |0xD0| Capital Eth
        'N',       # |0xD1| Capital N tilde
        'O',       # |0xD2| Accented capital O grave
        'O',       # |0xD3|                    acute
        'O',       # |0xD4|                    circumflex
        'O',       # |0xD5|                    tilde
        'O',       # |0xD6|                    diaeresis
        'x',       # |0xD7| Multiplication sign
        'O',       # |0xD8| Capital O slash
        'U',       # |0xD9| Accented capital U grave
        'U',       # |0xDA|                    acute
        'U',       # |0xDB|                    circumflex
        'U',       # |0xDC|                    diaeresis
        'Y',       # |0xDD| Capital Y acute
        'Th',      # |0xDE| Capital thorn
        'ss',      # |0xDF| German small sharp s
        'a',       # |0xE0| Accented small a grave
        'a',       # |0xE1|                  acute
        'a',       # |0xE2|                  circumflex
        'a',       # |0xE3|                  tilde
        'a',       # |0xE4|                  diaeresis
        'a',       # |0xE5| Small a ring
        'ae',      # |0xE6| Small ae
        'c',       # |0xE7| Small c cedilla
        'e',       # |0xE8| Accented small e grave
        'e',       # |0xE9|                  acute
        'e',       # |0xEA|                  circumflex
        'e',       # |0xEB|                  diaeresis
        'i',       # |0xEC| Accented small i grave
        'i',       # |0xED|                  acute
        'i',       # |0xEE|                  circumflex
        'i',       # |0xEF|                  diaeresis
        'th',      # |0xF0| Small eth
        'n',       # |0xF1| Small n tilde
        'o',       # |0xF2| Accented small o grave
        'o',       # |0xF3|                  acute
        'o',       # |0xF4|                  circumflex
        'o',       # |0xF5|                  tilde
        'o',       # |0xF6|                  diaeresis
        '/',       # |0xF7| Division sign
        'o',       # |0xF8| Small o slash
        'u',       # |0xF9| Accented small u grave
        'u',       # |0xFA|                  acute
        'u',       # |0xFB|                  circumflex
        'u',       # |0xFC|                  diaeresis
        'y',       # |0xFD| Small y acute
        'th',      # |0xFE| Small thorn
        'y'        # |0xFF| Small y diaeresis
    );

    my (%topics, %authors);

    my ($yearstart, $yearend, $monthfirst, $monthlast) = (-1, -1, -1. -1);

    my ($currentOutputFile, $currentBackupFile);

    my $lastdate = '0000-00-00';

    my $plinkno = 0;
    my $lineno = 0;
    while ($l = <>) {
        $l =~ s/\s+$//;
        $lineno++;

        if (!($l =~ m/^(\w+):\s*(.*$)/)) {
            next;
        }
        my ($item, $arg) = ($1, $2);

        if ($item eq 'TITLE') {
            $title = $arg;
        } elsif ($item eq 'CATEGORY') {
            $category = $arg;
        } elsif ($item eq 'TOPIC') {
            $topics{$arg}++;
            if ($topic ne '') {
                $topic .= '::';
            }
            $topic .= $arg;
        } elsif ($item eq 'DATE') {
            $date = $arg;
            if ($date !~ m/^\d\d\d\d\-\d\d\-\d\d$/) {
                die("Malformed date ($date) on line $lineno");
            }
            if ($date lt $lastdate) {
                die("Date sequence error ($date) after ($lastdate) on line $lineno");
            }
            $lastdate = $date;
            $date =~ m/(\d+)\-(\d+)\-/;
            my ($year, $month) = ($1, $2);
            if ($yearstart < 0) {
                $yearstart = $year;
                $monthfirst = $month;
            }
            $yearend = $year;
            $monthlast = $month;
        } elsif ($item eq 'AUTHOR') {
            $authors{$arg}++;
            if ($author ne '') {
                $author .= '::';
            }
            $author .= $arg;
        } elsif ($item eq 'BODY') {
            while ($l = <>) {
                $l =~ s/\s+$//;
                if ($l eq '--------') {
                    last;
                }
                if ($l ne '') {
                    $body .= "$l\n";
                }
            }
            if ($body !~ m/<dd>/) {
                $body .= "<dd>\n";
            }
            $plinkno++;
            $body = expandMacros($body, $plinkno);
            push(@books, { title => $title,
                           category => $category,
                           topic => $topic,
                           date => $date,
                           author => $author,
                           permalink => "book_$plinkno.html",
                           body => $body
                         });
            ($title, $category, $topic, $date, $author, $body) = ('', '', '', '', '', '');
        }
    }

    my (@topics, @authors);

    foreach $l (sort { lc(isoflat($a)) cmp lc(isoflat($b)) } keys %topics) {
        push(@topics, $l);
    }

    foreach $l (sort { lc(isoflat($a)) cmp lc(isoflat($b)) } keys %authors) {
        push(@authors, $l);
    }

    print(STDERR "Books: $#books  Authors: $#authors  Topics: $#topics\n");

    my $lonely = '';
    foreach $l (sort { lc(isoflat($a)) cmp lc(isoflat($b)) } keys %topics) {
        if ($topics{$l} == 1) {
            if ($lonely eq '') {
                $lonely = "Lonely topics: $l";
            } else {
                $lonely .= ", $l";
            }
        }
    }
    if ($lonely ne '') {
        print(STDERR "$lonely\n");
    }

    #   Generate the global navigation panel.  The panels named
    #   topicsN.html are clones of the "Welcome" panel, but have
    #   the respective button highlighted.

    open(FIDX, ">html/topics.html");
    navpanel(0, 0);
    close(FIDX);

    foreach $l (1, 2, 7, 8) {
        open(FIDX, ">html/topics$l.html");
        navpanel(0, $l);
        close(FIDX);
    }

    #   Generate no frames document

    {
        print(STDERR "Generating no frames document.\n");
        my ($lastyear, $lastmonth) = (-1, -1);
        my $intab = 0;
        open(FI, "<noframes.htmm");
        openOut("html/noframes.html");

        while ($l = <FI>) {
            if ($l =~ m/^<!\-\-#include file="books\.htmi" \-\->/) {
                last;
            }
            print(FO $l);
        }

        $books[$#books]{date} =~ m/(\d+)\-(\d+)\-/;
        print(FO "<h4><a href=\"#$1-$2\">Most Recent Additions</a></h4>\n");
        print(FO "<h4><a href=\"../\">Frames Edition</a></h4>\n");

        for (my $i = 0; $i <= $#books; $i++) {
            my $d = $books[$i]{date};
            $d =~ m/(\d+)\-(\d+)\-/;
            my ($year, $month) = ($1, $2);
            if ($lastyear != $year) {
                if ($intab) {
                    print(FO "</dl>\n");
                    $intab = 0;
                }
                yearhead($year);
                $lastyear = $year;
                $lastmonth = -1;
            }
            if ($lastmonth != $month) {
                if ($intab) {
                    print(FO "</dl>\n");
                    $intab = 0;
                }
                monthhead($year, $month);
                $lastmonth = $month;
            }
            if (!$intab) {
                print(FO "<dl>\n");
                $intab = 1;
            }
            my $b = $books[$i]{body};
            #   Transform links to months to internal links within the no-frames document
            $b =~ s|<a onclick="parent\.elements\.location\.href='idx_month\.html#(\d+)';"\nhref="month_(\d+)\-(\d+)\.html">|<a href="#$2-$3">|g;
            print(FO $b);
            permalink($books[$i]{permalink});
        }

        if ($intab) {
            print(FO "</dl>\n");
        }
        while ($l = <FI>) {
            print(FO $l);
        }
        close(FI);
        closeOut();
    }

    #   Generate most recent books chronological document and
    #   RSS 1.0 update feed.

    {
        print(STDERR "Generating most recent books chronological document.\n");
        my ($lastyear, $lastmonth) = (-1, -1);
        my $intab = 0;
        my $nbooks = 10;

        open(FR, ">html/index.rdf");
        print FR << "EOF";
<?xml version="1.0" encoding="iso-8859-1"?>

<rdf:RDF
xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
xmlns:dc="http://purl.org/dc/elements/1.1/"
xmlns:sy="http://purl.org/rss/1.0/modules/syndication/"
xmlns:admin="http://webns.net/mvcb/"
xmlns="http://purl.org/rss/1.0/">

<channel rdf:about="$baseurl/">
<title>${reader}'s Reading List</title>
<link>$baseurl/</link>
<description>$reader&apos;s Reading List</description>
<sy:updatePeriod>daily</sy:updatePeriod>
<sy:updateFrequency>1</sy:updateFrequency>
<sy:updateBase>2004-11-01T00:00:00Z</sy:updateBase>
<dc:language>en-gb</dc:language>
<dc:creator></dc:creator>
EOF

        my @now = gmtime(time());
        print(FR sprintf("<dc:date>%4d-%02d-%02d</dc:date>\n\n<items>\n<rdf:Seq>\n",
                $now[5] + 1900, $now[4] + 1, $now[3]));
        for (my $i = $#books; ($i >= 0) && ($i > ($#books - $nbooks)); $i--) {
            print(FR "<rdf:li rdf:resource=\"$baseurl/indices/$books[$i]{permalink}\" />\n");
        }
        print(FR "</rdf:Seq>\n</items>\n\n</channel>\n");

        for (my $i = $#books; ($i >= 0) && ($i > ($#books - $nbooks)); $i--) {
            my $au = $books[$i]{author};
            $au .= '::';
            my $a = '';
            while ($au ne '') {
                if ($a ne '') {
                    $a .= ', ';
                }
                $au =~ m/^([^\,\:]+)/;
                $a .= $1;
                $au =~ s/^[^\:]*\:\://;
            }

            my $abb = $books[$i]{body};
            $abb =~ s/<dt>//;
            $abb =~ s/<dd>.*$//s;
            $abb =~ s/\n/ /gs;
            $abb =~ s/\s+/ /g;
            $abb =~ s/&nbsp;/ /g;
            $abb = stripHTML($abb);
            print FR << "EOF";

<item rdf:about="$baseurl/indices/$books[$i]{permalink}">
<title><![CDATA[$books[$i]{title} ($a)]]></title>
<link>$baseurl/indices/$books[$i]{permalink}</link>
<description><![CDATA[$abb]]></description>
<dc:subject>Reading List</dc:subject>
<dc:creator>$reader</dc:creator>
<dc:date>$books[$i]{date}</dc:date>
</item>
EOF
        }

        print(FR "</rdf:RDF>\n");
        close(FR);

        openOut("html/recent.html");
        htmlhead("Most Recent Books Newest to Oldest");

        print(FO "<h2 align=\"center\">Last $nbooks Books Read, Newest to Oldest</h2>\n");
        print(FO "<h4 align=\"center\"><a href=\"../rindex.html\" target=\"_top\">Bookmark this link to come directly here.</a></h4>\n");

        for (my $i = $#books; ($i >= 0) && ($i > ($#books - $nbooks)); $i--) {
            my $d = $books[$i]{date};
            $d =~ m/(\d+)\-(\d+)\-/;
            my ($year, $month) = ($1, $2);
            if ($lastyear != $year) {
                $lastyear = $year;
                $lastmonth = -1;
            }
            if ($lastmonth != $month) {
                if ($intab) {
                    print(FO "</dl>\n");
                    $intab = 0;
                }
                monthhead($year, $month);
                $lastmonth = $month;
            }
            if (!$intab) {
                print(FO "<dl>\n");
                $intab = 1;
            }
            print(FO $books[$i]{body});
            permalink($books[$i]{permalink});
        }

        if ($intab) {
            print(FO "</dl>\n");
        }
        htmltail();
        closeOut();
    }

    #   Generate "weblog.txt" document in current directory
    #   with the most recent entry in format suitable for
    #   pasting into the update log.

    {

        print(STDERR "Generating Web log entry for newest book.\n");
        my $i = $#books;
        openOut("weblog.txt");
        my $t = uninvert($books[$i]{title});
        print(FO "Reading List: <cite>$t</cite>\n");
        print(FO "<dl>\n");
        my $b = $books[$i]{body};
        $b =~ s/\n\n/\n/g;
        #   Transform the HTML generated by <Month> macros into a link which displays the
        #   reading list document with the specified month in the main window.
        $b =~ s|<a onclick="parent\.elements\.location\.href='idx_month\.html#(\d+)';"\nhref="month_(\d+)\-(\d+)\.html">|<a href="/$webdir/?month=$2-$3" target="$auxwin">|g;
        #   Transform HTML generated by the <Book> macro to open the item in an
        #   auxiliary window.
        $b =~ s|($baseurl/\?book=\d+")>|$1\ntarget="$auxwin">|g;
        print(FO $b);
        print(FO "</dd>\n</dl>\n");
        closeOut();
    }


    #   Generate all books chronological document

    {
        print(STDERR "Generating all books chronological document.\n");
        my ($lastyear, $lastmonth) = (-1, -1);
        my $intab = 0;
        openOut("html/allbooks.html");
        htmlhead("All Books in Order Read");

        for (my $i = 0; $i <= $#books; $i++) {
            my $d = $books[$i]{date};
            $d =~ m/(\d+)\-(\d+)\-/;
            my ($year, $month) = ($1, $2);
            if ($lastyear != $year) {
                if ($intab) {
                    print(FO "</dl>\n");
                    $intab = 0;
                }
                yearhead($year);
                $lastyear = $year;
                $lastmonth = -1;
            }
            if ($lastmonth != $month) {
                if ($intab) {
                    print(FO "</dl>\n");
                    $intab = 0;
                }
                monthhead($year, $month);
                $lastmonth = $month;
            }
            if (!$intab) {
                print(FO "<dl>\n");
                $intab = 1;
            }
            print(FO $books[$i]{body});
            permalink($books[$i]{permalink});
        }

        if ($intab) {
            print(FO "</dl>\n");
        }
        htmltail();
        closeOut();
    }

    #   Generate yearly chronological documents

    print(STDERR "Generating yearly chronological documents.\n");
    open(FIDX, ">html/idx_year.html");
    navpanel(1, 3);
    for (my $y = $yearstart; $y <= $yearend; $y++) {
        my ($lastyear, $lastmonth) = (-1, -1);
        my $intab = 0;
        print FIDX << "EOF";
<tr><td><p class="i"><a class="i" href="year_$y.html">$y</a></p></td></tr>
EOF
        openOut("html/year_$y.html");
        htmlhead("Books read in $y");

        for (my $i = 0; $i <= $#books; $i++) {
            my $d = $books[$i]{date};
            $d =~ m/(\d+)\-(\d+)\-/;
            my ($year, $month) = ($1, $2);
            if ($year == $y) {
                if ($lastyear != $year) {
                    if ($intab) {
                        print(FO "</dl>\n");
                        $intab = 0;
                    }
                    yearhead($year);
                    $lastyear = $year;
                    $lastmonth = -1;
                }
                if ($lastmonth != $month) {
                    if ($intab) {
                        print(FO "</dl>\n");
                        $intab = 0;
                    }
                    monthhead($year, $month);
                    $lastmonth = $month;
                }
                if (!$intab) {
                    print(FO "<dl>\n");
                    $intab = 1;
                }
                print(FO $books[$i]{body});
                permalink($books[$i]{permalink});
            }
        }

        if ($intab) {
            print(FO "</dl>\n");
        }
        htmltail();
        closeOut();
    }
    navpanel(-1, 3);
    close(FIDX);

    #   Generate monthly chronological documents

    print(STDERR "Generating monthly chronological documents.\n");
    open(FIDX, ">html/idx_month.html");
    navpanel(2, 4);
    {
         my ($lastyear, $lastmonth) = (-1, -1);

        for (my $y = $yearstart; $y <= $yearend; $y++) {
            for (my $m = 1; $m <= 12; $m++) {
                my $intab = 0;
                my $sm = sprintf("%02d", $m);
                my $tm = $months[$m - 1];
                my $opened = 0;

                for (my $i = 0; $i <= $#books; $i++) {
                    my $d = $books[$i]{date};
                    $d =~ m/(\d+)\-(\d+)\-/;
                    my ($year, $month) = ($1, $2);
                    if (($year == $y) && ($month == $m)) {
                        if (!$opened) {
                            openOut("html/month_$y-$sm.html");
                            htmlhead("Books read in $tm $y");
                            $opened = 1;
                        }
                        if ($lastyear != $y) {
                            print FIDX << "EOF";
<tr><th><a class="n" name="$y">$y</a></th></tr>
EOF
                            if ($intab) {
                                print(FO "</dl>\n");
                                $intab = 0;
                            }
                            yearhead($year);
                            $lastyear = $year;
                            $lastmonth = -1;
                        }
                        if ($lastmonth != $month) {
                            print FIDX << "EOF";
<tr><td><p class="i"><a class="i" name="$y-$sm" href="month_$y-$sm.html">$tm</a></p></td></tr>
EOF
                            if ($intab) {
                                print(FO "</dl>\n");
                                $intab = 0;
                            }
                            monthhead($year, $month);
                            $lastmonth = $month;
                        }
                        if (!$intab) {
                            print(FO "<dl>\n");
                            $intab = 1;
                        }
                        print(FO $books[$i]{body});
                        permalink($books[$i]{permalink});
                    }
                }
                if ($opened) {
                    if ($intab) {
                        print(FO "</dl>\n");
                    }
                    htmltail();
                    closeOut();
                }
            }
        }
    }
    navpanel(-2, 4);
    close(FIDX);

    #   Generate list by topic

    {
        print(STDERR "Generating by topic lists.\n");
        my @tbooks = sort { lc(isoflat(stripHTML($a->{author}))) cmp lc(isoflat((stripHTML($b->{author})))) } @books;
        open(FIDX, ">html/idx_topic.html");
        navpanel(3, 5);
        for (my $i = 0; $i <= $#topics; $i++) {
            my $thist = $topics[$i];
            my $topfile = tofile($thist);
            my $topname = lc($topfile);
            openOut("html/topic_$topfile.html");
            print FIDX << "EOF";
<tr><td><p class="i"><a class="i" name="$topname" href="topic_$topfile.html">$thist</a></p></td></tr>
EOF
            htmlhead("Topic: $thist");
            print(FO "<h2 align=\"center\">$thist</h2>\n");
            print(FO "<dl>\n");
            my $ntopic = 0;
            for (my $j = 0; $j <= $#tbooks; $j++) {
                my $t = '::' . $tbooks[$j]{topic} . '::';
                if (index($t, "\:\:$thist\:\:") >= 0) {
                    print(FO $tbooks[$j]{body});
                    datelink($tbooks[$j]{date}, $tbooks[$j]{permalink});
                    $ntopic++;
                }
            }
            die("No books found for title ($thist)") if ($ntopic == 0);
            print(FO "</dl>\n");
            htmltail();
            closeOut();
        }
        navpanel(-3, 5);
        close(FIDX);
    }

    #   Generate list by author

    {
        print(STDERR "Generating by author lists.\n");
        my @abooks = sort { lc(isoflat(stripHTML($a->{title}))) cmp lc(isoflat((stripHTML($b->{title})))) } @books;
        open(FIDX, ">html/idx_author.html");
        navpanel(4, 6);
        my $lastletter = '';
        for (my $i = 0; $i <= $#authors; $i++) {
            my $thisa = $authors[$i];
            my $aufile = tofile($thisa);
            openOut("html/author_$aufile.html");
            my $ll = $thisa;
            $ll =  uc(isoflat((stripHTML($ll))));
            $ll =~ m/^(.)/;
            $ll = $1;
            if ($ll ne $lastletter) {
                my $lid = $ll;
                if ($lid !~ m/[A-Z\d]/) {
                    #   If first letter of author is non-alphanumeric, use hex for the ID
                    $lid = sprintf("x%X", ord($lid));
                }
                $lastletter = $ll;
                print(FIDX "<tr><th><a class=\"n\" id=\"Au_$lid\" name=\"au_$lid\">&ndash; $lastletter &ndash;</a></th></tr>\n");
            }
            print FIDX << "EOF";
<tr><td><p class="i"><a class="i" href="author_$aufile.html">$thisa</a></p></td></tr>
EOF
            htmlhead("Books by $thisa");
            print(FO "<h2 align=\"center\">Books by $thisa</h2>\n");
            print(FO "<dl>\n");
            my $nauth = 0;
            for (my $j = 0; $j <= $#abooks; $j++) {
                my $a = '::' . $abooks[$j]{author} . '::';
                if (index($a, "\:\:$thisa\:\:") >= 0) {
                    print(FO $abooks[$j]{body});
                    datelink($abooks[$j]{date}, $abooks[$j]{permalink});
                    $nauth++;
                }
            }
            die("No books found for author ($thisa)") if ($nauth == 0);
            print(FO "</dl>\n");
            htmltail();
            closeOut();
        }
        navpanel(-4, 6);
        close(FIDX);
    }

    #   Generate list by title

    {
        print(STDERR "Generating by title list.\n");
        openOut("html/idx_title.html");
        htmlhead("Books by Title", 'bytitle');
        my $lastletter = '';
        my @tbooks = sort { lc(isoflat(stripHTML($a->{title}))) cmp lc(isoflat((stripHTML($b->{title})))) } @books;
        for (my $i = 0; $i <= $#tbooks; $i++) {
            my $ll = $tbooks[$i]{title};
            $ll =  uc(isoflat((stripHTML($ll))));
            $ll =~ m/^(.)/;
            $ll = $1;
            if ($ll ne $lastletter) {
                $lastletter = $ll;
                my $lid = $lastletter;
                if ($lid !~ m/[A-Z\d]/) {
                    #   If first letter of author is non-alphanumeric, use hex for the ID
                    $lid = sprintf("x%X", ord($lid));
                }
                print(FO "<h1 align=\"center\"><a class=\"i\" name=\"ti_$lid\" id=\"Ti_$lid\">$lastletter</a></h1>\n");
            }
            my $au = $tbooks[$i]{author};
            $au .= '::';
            my $a = '';
            while ($au ne '') {
                if ($a ne '') {
                    $a .= ', ';
                }
                $au =~ m/^([^\,\:]+)/;
                $a .= $1;
                $au =~ s/^[^\:]*\:\://;
            }
            print(FO "<a href=\"$tbooks[$i]{permalink}\">$tbooks[$i]{title}</a> &nbsp; (<i>$a</i>)<br />\n");
        }
        htmltail();
        closeOut();

        #   Generate individual documents for books

        for (my $i = 0; $i <= $#tbooks; $i++) {
            openOut("html/$tbooks[$i]{permalink}");
            htmlhead(stripHTML($tbooks[$i]{title}), 'book', $tbooks[$i]);
            print(FO "<dl>\n");
            print(FO $tbooks[$i]{body});
            datelink($tbooks[$i]{date}, , $tbooks[$i]{permalink});
            print(FO "</dl>\n");
            my $ll = $tbooks[$i]{title};
            $ll =  uc(isoflat((stripHTML($ll))));
            $ll =~ m/^(.)/;
            $ll = $1;
            print(FO "<h4 class=\"nav\"><a href=\"../index.html?title=$ll\" target=\"_top\">Other Book Reviews at Fourmilab</a></h4>\n");
            print(FO "<h4 class=\"nav\"><a href=\"/\" target=\"_top\">Fourmilab Home Page</a></h4>\n");
            htmltail();
            closeOut();
        }
    }

    #   Generate HTML heading for book list

    sub htmlhead {
        my ($title, $type, $book) = @_;

        print FO << "EOF";
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
<head>
<title>$title</title>
<link rel="stylesheet" href="../booklist.css" type="text/css" />
<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" />
EOF

    my $bodySpec = "";

    if (defined($type)) {
        if ($type eq 'book') {
            my $author = $$book{author};
            #   Uninvert the title for the description and keywords
            my $t = uninvert($title);
            my @a = split(/::/, $author);
            #   Uninvert the author name(s)
            for (my $i = 0; $i <= $#a; $i++) {
                if ($a[$i] =~ s/,\s([^,]+)$//) {
                    $a[$i] = "$1 $a[$i]";
                }
            }
            #   Build a list of authors for the description item
            my $authors = join(", ", @a);
            if ($#a > 1) {
                $authors =~ s/,([^,]+)$/, and$1/;
            } else {
                $authors =~ s/,([^,]+)$/ and$1/;
            }
            #   Assemble the list of keywords from the title and author
            my $kw = "book review $title $authors fourmilab $reader";
            $kw =~ s/&\w+;/ /g;         # Convert XHTML text entities to white space
            $kw =~ s/\s+l'/ /ig;        # Remove "l'" from start of French words
            $kw =~ s/\s+Jr\.\s+/ /ig;   # Remove "Jr."
            $kw =~ s/\-/ /g;            # Separate compound words
            $kw =~ s/[^\w\s'\xC0-\xD6\xD8-\xF6\xF8-\xFF]//g;    # Remove remaining punctuation
            $kw =~ s/\s+\w\s+/ /g;      # Collapse single-letter words
                # Delete common words
            while ($kw =~ s/\s+(and|or|the|in|of|with|for|to|le|la|les|et|ou||au|de|du)\s+/ /i) {} ;
            $kw =~ s/\s+/,/g;           # Separate keywords with commas
            $kw =~ s/(,[^,]+)(,.*)(\1,)/$1$2,/g;     # Elide duplicate keywords
            print FO << "EOF";
<meta name="description" content="Book review: $t by $authors" />
<meta name="keywords" content="$kw" />
<meta name="robots" content="index" />
EOF
        } elsif ($type eq 'bytitle') {
            $bodySpec = " onkeydown=\"TitleNav(event);\"";
            print FO << "EOF";
<meta name="description" content="$title" />
<meta name="robots" content="noindex,follow" />
EOF
        }
    } else {
        print FO << "EOF";
<meta name="description" content="$title" />
<meta name="robots" content="noindex,nofollow" />
EOF
    }

    print FO << "EOF";
<meta name="author" content="$reader" />

<script type="text/javascript" language="JavaScript" src="../booklist.js" defer="defer">
</script>

</head>

<body bgcolor="#FFFFFF" onunload="unloading();"$bodySpec>

<div class="bodycopy">
EOF
    }

    #   Generate HTML trailer for book list

    sub htmltail {
        print FO << "EOF";
</div>

</body>
</html>
EOF
    }

    #   Generate heading for year

    sub yearhead {
        my ($year) = @_;

        #   Backgrounds use resistor colour code of the last digit of the year
        my @colours = ( "#000000", "#996633", "#FF0000", "#FF6600", "#FFFF00",
                        "#33CC33", "#6666FF", "#CC66FF", "#939393", "#FFFFFF" );
        my $yearunit = $year % 10;
        my $col = $colours[$yearunit];
        my $border = ($yearunit == 9) ? ' style="border: 2px solid;"' : '';
        my $tcol = ($yearunit == 9) ? '#000000' : '#FFFFFF';
        print FO << "EOF";
<center>
<table cellpadding="8">
<tr><td bgcolor="$col"$border><font size="5" color="$tcol" face="Helvetica,Arial,sans-serif"><a class="i" name="$year">&nbsp; $year &nbsp;</a></font></td></tr>
</table>
</center>
EOF
    }

    #   Generate heading for month

    sub monthhead {
        my ($year, $month) = @_;

        my $mname = $months[$month - 1];
        print FO << "EOF";
<h3><a class="i" name="$year-$month">$mname $year</a></h3>
EOF
    }

    #   Generate left side navigation panel with plugged-in items
    #   The $which argument determines the point at which transcription
    #   will be paused to insert specific navigation items.  If 0, no
    #   pause is made--this creates the main navigation panel.  If
    #   positive, the panel is emitted up to and including that item,
    #   then paused.  If negative, the balance of the panel is generated
    #   after items have been interpolated.  The panel is written to file
    #   handle FIDX.

    sub navpanel {
        my ($which, $highlight) = @_;

        my @borh = qw(b b b b b b b b b b);
        $borh[$highlight] = 'h';
        
        my $bodySpec = "";
        
        if ($which == 4) {
            $bodySpec = " onkeydown=\"AuthorNav(event);\"";
        }

        if ($which >= 0) {
            print FIDX << "EOF";
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
<head>
<title>Fourmilab Topics</title>
<link rel="stylesheet" href="../navpanel.css" type="text/css" />
<base target="topic" />
<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" />
<meta name="author" content="$reader" />
<meta name="robots" content="noindex,follow" />
<script type="text/javascript" language="JavaScript" src="../navpanel.js" defer="defer">
</script>
</head>
<body bgcolor="#FFFFFF"$bodySpec>
<table width="100%" cellpadding="0">

<tr class="$borh[0]"><td class="b" align="center"><a class="s" accesskey="w" onclick="location.href='topics.html';" href="../welcome.html"><u>W</u>elcome</a></td></tr>
<tr><td class="filler"></td></tr>

<tr class="$borh[1]"><td class="b" align="center"><a class="s" accesskey="r" onclick="location.href='topics1.html';" href="recent.html">Most <u>R</u>ecent</a></td></tr>
<tr><td class="filler"></td></tr>

<tr class="$borh[2]"><td class="b" align="center"><a class="s" accesskey="a" onclick="location.href='topics2.html';" href="allbooks.html"><u>A</u>ll Books</a></td></tr>
<tr><td class="filler"></td></tr>

<tr class="$borh[3]"><td class="b" align="center"><a class="s" accesskey="y" href="idx_year.html" target="elements"><u>Y</u>early Lists</a></td></tr>
<tr><td class="filler"></td></tr>
EOF
        }

        if ($which == 1) {              # 1: Interpolate after Yearly Lists
            return;
        }

        if (($which >= 0) || (($which < 0) && ($which >= -1))) {
            print FIDX << "EOF";

<tr class="$borh[4]"><td class="b" align="center"><a class="s" accesskey="m" href="idx_month.html" target="elements"><u>M</u>onthly Lists</a></td></tr>
<tr><td class="filler"></td></tr>
EOF
        }

        if ($which == 2) {              # 2: Interpolate after Monthly Lists
            return;
        }

        if (($which >= 0) || (($which < 0) && ($which >= -2))) {
            print FIDX << "EOF";

<tr class="$borh[5]"><td class="b" align="center"><a class="s" accesskey="t" href="idx_topic.html" target="elements">By <u>T</u>opic</a></td></tr>
<tr><td class="filler"></td></tr>
EOF
        }

        if ($which == 3) {              # 3: Interpolate after Topics
            return;
        }

        if (($which >= 0) || (($which < 0) && ($which >= -3))) {
            print FIDX << "EOF";

<tr class="$borh[6]"><td class="b" align="center"><a class="s" accesskey="u" href="idx_author.html" target="elements">By A<u>u</u>thor</a></td></tr>
<tr><td class="filler"></td></tr>
EOF
        }

        if ($which == 4) {              # 4: Interpolate after Authors
            return;
        }

        print FIDX << "EOF";

<tr class="$borh[7]"><td class="b" align="center"><a class="s" accesskey="i" onclick="location.href='topics7.html';" href="idx_title.html">By T<u>i</u>tle</a></td></tr>
<tr><td class="filler"></td></tr>

<tr class="$borh[8]"><td class="b" align="center"><a class="s" accesskey="b" onclick="location.href='topics8.html';" href="../recommend.html">Recommend <u>B</u>ook</a></td></tr>
<tr><td class="filler"></td></tr>

<tr class="$borh[9]"><td class="b" align="center"><a class="s" accesskey="n" href="noframes.html" target="_top"><u>N</u>o Frames</a></td></tr>

<tr><td align="center">
&nbsp;<br />
<form name="feedback" method="post" action="/cgi-bin/FeedbackForm.pl" target="_top">
<input type="hidden" name="pagetitle" value="&lt;cite&gt;${reader}'s Reading List&lt;/cite&gt;" />
<input type="hidden" name="backlink" value="Back to &lt;cite&gt;${reader}'s Reading List&lt;/cite&gt;" />
<input type="hidden" name="return" value="/$webdir/" />
<input type="submit" accesskey="f" value=" Send Feedback " />
</form>
</td></tr>

<tr><td align="center">
<a href="index.rdf" type="application/rss+xml"><img src="../figures/rss10.png"
    width="80" height="15" border="0" alt="RSS 1.0" /></a>
<a href="http://validator.w3.org/feed/check.cgi?url=$baseurl/indices/index.rdf"
   target="Fourmilab_Reading_List_Validator"><img src="../figures/check.png"
    width="15" height="15" border="0" alt="Validate RSS 1.0 Feed" /></a>
</td></tr>

<tr><td align="center">
&nbsp;<br />
<table bgcolor="#D0D0D0" width="170" cellspacing="4">
<tr>
<td align="center">
<font face="Helvetica,Arial,sans-serif" size="-1">
Click on titles to order books on-line from
<br /><a href="http://www.amazon.com/?tag=$amazon_associate"
    target="Amazon_Fourmilab">
<img src="../figures/amazon.gif"
    width="90" height="29" border="0" alt="Amazon.com" /></a>
</font>
</td></tr>
</table>
</td></tr>

<tr><td align="center">
&nbsp;<br />
<table bgcolor="#D0D0D0" width="170" cellspacing="4">
<tr>
<td align="center">
<font face="Helvetica,Arial,sans-serif" size="-1">
Click on titles to find out of print books at
<br /><a href="http://www.abebooks.com/"
    target="Amazon_Fourmilab">
<img src="../figures/abe_books.gif"
    width="125" height="125" border="0" alt="abebooks.com" /></a>
</font>
</td></tr>
</table>
</td></tr>
</table>

<p />
<center>
<a href="http://validator.w3.org/check?uri=referer"
   target="Fourmilab_Reading_List_Validator"><img
   src="../figures/valid-xhtml10.png"
   alt="Valid XHTML 1.0" height="31" width="88"
   onclick="validate(); return false;"
   border="0" /></a>
<br />
<a href="http://jigsaw.w3.org/css-validator/validator?uri=$baseurl/booklist.css"
   target="Fourmilab_Reading_List_Validator"><img
      src="../figures/vcss.png"
      alt="Valid CSS"  height="31" width="88"
    border="0"/></a>
</center>

</body>
</html>
EOF
    }

    #   Generate date read reference for book with link to month index.
    #   Called with the ISO date for the book item and name of the
    #   permalink for the single book item.

    sub datelink {
        my ($d, $p) = @_;
        $d =~ m/(\d+)\-(\d+)\-/;
        my ($year, $month) = ($1, $2);
        my $tmonth = $months[$month - 1];
        my $perma = "&nbsp;<a href=\"$p\"><img src=\"../figures/link.png\" width=\"13\" height=\"9\" border=\"0\" alt=\"Permalink\" /></a>";
        print(FO "<p class=\"d\"><a onclick=\"parent.elements.location.href='idx_month.html';\" href=\"month_$year-$month.html\">$tmonth $year</a>$perma</p>\n");
        print(FO "</dd>\n");
    }

    #   Generate the permalink for a book item

    sub permalink {
        my ($p) = @_;
        my $perma = "&nbsp;<a href=\"$p\"><img src=\"../figures/link.png\" width=\"13\" height=\"9\" border=\"0\" alt=\"Permalink\" /></a>";
        print(FO "<p class=\"d\">$perma</p>\n");
        print(FO "</dd>\n");
    }


    #   Flatten ISO characters in a string to closest ASCII representation

    sub isoflat {
        my ($s) = @_;

        if ($s =~ m/[\xA0-\xFF]/) {
            $s =~ s/([\xA0-\xFF])/$flattenISO[ord($1)- 0xA0]/eg;
        }
        return $s;
    }

    #   Turn text string into an acceptable file name

    sub tofile {
        my ($s) = @_;

        $s = lc(isoflat($s));       # Flatten ISO and convert to lower case
        $s =~ s/[^\w\s-]//g;        # Elide non-word characters
        $s =~ s/\s+/_/g;            # Convert sequences of white space to underscores
        return $s;
    }


    #   Attempt to un-invert the title if it looks inverted (brutal hack)

    sub uninvert {
        my ($t) = @_;
        if ($t =~ s/,\s(\w+)$//) {
            $t = "$1 $t";
        }
        return $t;
    }

    #   Strip HTML (primitive, but good enough for our needs

    sub stripHTML {
        my ($s) = @_;

        if ($s =~ s/<[^>]*>//g) {
            $s =~ s/^\s+//;
            $s =~ s/\s$//g;
        }
        return $s;
    }

    #   Encode a URI with special characters in hexadecimal just as in JavaScript

    sub encodeURI {
        my ($s) = @_;

        $s =~ s/([^a-zA-Z0-9_.~-])/uc(sprintf("%%%02x", ord($1)))/eg;
        return $s;
    }

    #   Expand ampersands in Elink URLs to XHTML entities

    sub expandElink {
        my ($s) = @_;

        $s =~ s/&/&amp;/g;
        return "<a href=\"$s\"\ntarget=\"$auxwin\">";
    }

    #   Expand macros in body text

    sub expandMacros {
        my ($s, $plinkno) = @_;

        my $spoilno = 0;

        #   <Xlink url>         Link to external site
        $s =~ s|<Xlink\s+([^>]+)>|<a href="$1"\ntarget="$auxwin">|g;

        #   <Elink url>         Link to external site with & expanded to &amp;
        $s =~ s|<Elink\s+([^>]+)>|expandElink($1)|eg;

        #   <Amazon isbn>       Link to amazon.com
        $s =~ s|<Amazon\s+(\w+)>|<a href="http://www.amazon.com/dp/$1/?tag=$amazon_associate"\ntarget="Amazon_Fourmilab">|g;
        #   <Amazon.tld isbn>   Link to amazon.tld
        $s =~ s|<Amazon\.([\w\.]+)\s+(\w+)>|<a href="http://www.amazon.$1/dp/$2/?tag=$amazon_associate"\ntarget="Amazon_Fourmilab">|g;
        #   <Amspecial url>     Special link to amazon.com (series, listmania, etc.)
        $s =~ s|<Amspecial\s+([^>]+)>|"<a href=\"http://www.amazon.com/$1" .
                (($1 =~ m/\?/) ? "&amp;" : "/?") . "tag=$amazon_associate\"\ntarget=\"Amazon_Fourmilab\">"|eg;
        #   <Amspecial.tld url> Special link to amazon.tld (search, series, listmania, etc.)
        $s =~ s|<Amspecial\.([\w\.]+)\s+([^>]+)>|"<a href=\"http://www.amazon.$1/$2" .
                (($2 =~ m/\?/) ? "&amp;" : "/?") . "tag=$amazon_associate\"\ntarget=\"Amazon_Fourmilab\">"|eg;

        #   <Abe search_string> Search on abebooks.com
        $s =~ s|<Abe\s+([^>]+)>|"<a href=\"http://www.abebooks.com/servlet/SearchResults?" .
                $1 . "\"\ntarget=\"Amazon_Fourmilab\">"|eg;

        #   <Medium text>
        $s =~ s|<Medium\s+([^>]+)>|<span class="medium">\($1\).</span>|g;

        #   <Audiobook>         Audio book icon
        $s =~ s|<Audiobook>|<img src="/$webdir/figures/audiobook.png" width="17" height="17" alt="[Audiobook]" />|g;
        #   <Audible productno> Link to audible.com
        $s =~ s|<Audible\s+(\w+)>|<a href="http://www.audible.com/adbl/store/welcome.jsp\?entryRedirect=/site/products/ProductDetail.jsp&amp;entryParams=^productID~$1"\ntarget="Amazon_Fourmilab">|g;

        #   <Book nnn>
        $s =~ s|<Book (\d+)>|<a href="$baseurl/?book=$1" target="_top">|g;

        #   <Month yyyy-mm>
        $s =~ s|<Month (\d+)\-(\d+)>|"<a onclick=\"parent.elements.location.href='idx_month.html#$1';\"\nhref=\"month_" .
                sprintf("%d-%02d", $1, $2) . ".html\">" . $months[$2 - 1] . "&nbsp;$1</a>"|eg;

        #   <Dict>word or phrase</Dict>
        $s =~ s|<Dict>([^<]+)</Dict>|<a href="http://dictionary.reference.com/browse/$1" target="$auxwin">$1</a>|g;

        #   <Spoiler> and <Endspoiler>
        $s =~ s|<Spoiler>|"<div class=\"spoiler\">Spoiler warning: <em>Plot and/or ending details follow.</em> &nbsp;
    <span class=\"trigger\" style=\"display: none;\" id=\"show_spoilers_${plinkno}_" . (++$spoilno) . "\">
    <a href=\"#\"
        onclick=\"document.getElementById('spoilers_${plinkno}_$spoilno').style.display='block'; document.getElementById('hide_spoilers_${plinkno}_$spoilno').style.display='inline'; document.getElementById('show_spoilers_${plinkno}_$spoilno').style.display='none'; return false;\">(Show
        Spoilers)</a>
    <script type=\"text/javascript\">
    /* <![CDATA[ */
        document.getElementById(\"show_spoilers_${plinkno}_$spoilno\").style.display = \"inline\";
    /* ]]> */
    </script>
    </span>
    <span class=\"trigger\" style=\"display: none;\" id=\"hide_spoilers_${plinkno}_$spoilno\">
        <a href=\"#\" onclick=\"document.getElementById('show_spoilers_${plinkno}_$spoilno').style.display='inline'; document.getElementById('hide_spoilers_${plinkno}_$spoilno').style.display = 'none'; document.getElementById('spoilers_${plinkno}_$spoilno').style.display = 'none'; return false;\">(Hide
        Spoilers)</a>
    </span>
</div>
    <div class=\"spoilers\" style=\"display: block;\" id=\"spoilers_${plinkno}_$spoilno\">
    <script type=\"text/javascript\">
    /* <![CDATA[ */
        document.getElementById(\"spoilers_${plinkno}_$spoilno\").style.display = \"none\";
    /* ]]> */
    </script>"|eg;

        $s =~ s|<Endspoiler>|<div class="endspoiler"><em>Spoilers end here.</em> &nbsp;
    <a href="#" style=\"display: none;\" id=\"hide_spoilers_e_${plinkno}_$spoilno\" onclick="document.getElementById('show_spoilers_${plinkno}_$spoilno').style.display='inline'; document.getElementById('hide_spoilers_${plinkno}_$spoilno').style.display = 'none'; document.getElementById('spoilers_${plinkno}_$spoilno').style.display = 'none'; return false;">(Hide
            Spoilers)</a>
    <script type=\"text/javascript\">
    /* <![CDATA[ */
        document.getElementById(\"hide_spoilers_e_${plinkno}_$spoilno\").style.display = \"inline\";
    /* ]]> */
    </script>
    </div>
</div>|g;

        return $s;
    }

    #   Start writing an output file

    sub openOut {
        ($currentOutputFile) = @_;
        $currentBackupFile = '';
        if (-f $currentOutputFile) {
            $currentBackupFile = "$currentOutputFile.BAK";
            system("mv $currentOutputFile $currentBackupFile");
        }
        open(FO, ">$currentOutputFile") || die("Cannot open $currentOutputFile for output");
    }

    #   End writing an output file.  If the content is
    #   unchanged from the previous version, restore the
    #   original file to minimise files we need propagate
    #   across the server farm with rdist.

    sub closeOut {
        close(FO);
        if ($currentBackupFile ne '') {
            if ((system("cmp -s $currentBackupFile $currentOutputFile") >> 8) == 0) {
                system("mv $currentBackupFile $currentOutputFile");
#print(STDERR "Reverting $currentBackupFile to $currentOutputFile: files identical\n");
            } else {
                system("rm $currentBackupFile\n");
print(STDERR "    Updated $currentOutputFile\n");
            }
        } else {
print(STDERR "    Created $currentOutputFile\n");
        }
    }
